import argparse
import inspect
from copy import deepcopy
from functools import partial, update_wrapper
from typing import List, Mapping, Sequence, _GenericAlias

import yaml


class Registrable:
    """Class used to denote which types of objects can be registered in the RLHive
    Registry. These objects can also be configured directly from the command line, and
    recursively built from the config, assuming type annotations are present.
    """

    @classmethod
    def type_name(cls):
        """This should represent a string that denotes the which type of class you are
        creating. For example, "logger", "agent", or "env".
        """
        raise ValueError


class CallableType(Registrable):
    """A wrapper that allows any callable to be registered in the RLHive Registry.
    Specifically, it maps the arguments and annotations of the wrapped function to the
    resulting callable, allowing any argument names and type annotations of the
    underlying function to be present for outer wrapper. When called with some
    arguments, this object returns a partial function with those arguments assigned.

    By default, the type_name is "callable", but if you want to create specific types
    of callables, you can simply create a subclass and override the type_name method.
    See :py:class:`hive.utils.utils.OptimizerFn`.
    """

    def __init__(self, fn):
        """
        Args:
            fn: callable to be wrapped.
        """
        self._fn = fn
        update_wrapper(self, self._fn)

    def __call__(self, *args, **kwargs):
        return partial(self._fn, *args, **kwargs)

    @classmethod
    def type_name(cls):
        return "callable"

    def __repr__(self):
        return f"<{type(self).__name__} {repr(self._fn)}>"


class Registry:
    """This is the Registry class for RLHive. It allows you to register different types
    of :py:class:`Registrable` classes and objects and generates constructors for those
    classes in the form of `get_{type_name}`.

    These constructors allow you to construct objects from dictionary configs. These
    configs should have two fields: `name`, which corresponds to the name used when
    registering a class in the registry, and `kwargs`, which corresponds to the keyword
    arguments that will be passed to the constructor of the object. These constructors
    can also build objects recursively, i.e. if a config contains the config for
    another `Registrable` object, this will be automatically created before being
    passed to the constructor of the original object. These constructors also allow you
    to directly specify/override arguments for object constructors directly from the
    command line. These parameters are specified in dot notation. They also are able
    to handle lists and dictionaries of Registrable objects.

    For example, let's consider the following scenario:
    Your agent class has an argument `arg1` which is annotated to be `List[Class1]`,
    `Class1` is `Registrable`, and the `Class1` constructor takes an argument `arg2`.
    In the passed yml config, there are two different Class1 object configs listed.
    the constructor will check to see if both `--agent.arg1.0.arg2` and
    `--agent.arg1.1.arg2` have been passed.

    The parameters passed in the command line will be parsed according to the type
    annotation of the corresponding low level constructor. If it is not one of
    `int`, `float`, `str`, or `bool`, it simply loads the string into python using a
    yaml loader.

    Each constructor returns the object, as well a dictionary config with all the
    parameters used to create the object and any Registrable objects created in the
    process of creating this object.
    """

    def __init__(self) -> None:
        self._registry = {}

    def register(self, name, constructor, type):
        """Register a Registrable class/object with RLHive.

        Args:
            name (str): Name of the class/object being registered.
            constructor (callable): Callable that will be passed all kwargs from
                configs and be analyzed to get type annotations.
            type (type): Type of class/object being registered. Should be subclass of
                Registrable.

        """
        if not issubclass(type, Registrable):
            raise ValueError(f"{type} is not Registrable")
        if type.type_name() not in self._registry:
            self._registry[type.type_name()] = {}

            def getter(self, object_or_config, prefix=None):
                if object_or_config is None:
                    return None, {}
                elif isinstance(object_or_config, type):
                    return object_or_config, {}
                name = object_or_config["name"]
                kwargs = object_or_config.get("kwargs", {})
                expanded_config = deepcopy(object_or_config)
                if name in self._registry[type.type_name()]:
                    object_class = self._registry[type.type_name()][name]
                    parsed_args = get_callable_parsed_args(object_class, prefix=prefix)
                    kwargs.update(parsed_args)
                    kwargs, kwargs_config = construct_objects(
                        object_class, kwargs, prefix
                    )
                    expanded_config["kwargs"] = kwargs_config
                    return object_class(**kwargs), expanded_config
                else:
                    raise ValueError(f"{name} class not found")

            setattr(self.__class__, f"get_{type.type_name()}", getter)
        self._registry[type.type_name()][name] = constructor

    def register_all(self, base_class, class_dict):
        """Bulk register function.

        Args:
            base_class (type): Corresponds to the `type` of the register function
            class_dict (dict[str, callable]): A dictionary mapping from name to
                constructor.
        """
        for cls in class_dict:
            self.register(cls, class_dict[cls], base_class)

    def __repr__(self):
        return str(self._registry)


def construct_objects(object_constructor, config, prefix=None):
    """Helper function that constructs any objects specified in the config that
    are registrable.

    Returns the object, as well a dictionary config with all the parameters used to
    create the object and any Registrable objects created in the process of creating
    this object.

    Args:
        object_constructor (callable): constructor of object that corresponds to
            config. The signature of this function will be analyzed to see if there
            are any :py:class:`Registrable` objects that might be specified in the
            config.
        config (dict): The kwargs for the object being created. May contain configs for
            other `Registrable` objects that need to be recursively created.
        prefix (str): Prefix that is attached to the argument names when looking for
            command line arguments.
    """
    signature = inspect.signature(object_constructor)
    prefix = "" if prefix is None else f"{prefix}."
    expanded_config = deepcopy(config)
    for argument in signature.parameters:
        if argument not in config:
            continue
        expected_type = signature.parameters[argument].annotation

        if isinstance(expected_type, type) and issubclass(expected_type, Registrable):
            config[argument], expanded_config[argument] = registry.__getattribute__(
                f"get_{expected_type.type_name()}"
            )(config[argument], f"{prefix}{argument}")
        if isinstance(expected_type, _GenericAlias):
            origin = expected_type.__origin__
            args = expected_type.__args__
            if (
                (origin == List or origin == list)
                and len(args) == 1
                and isinstance(args[0], type)
                and issubclass(args[0], Registrable)
                and isinstance(config[argument], Sequence)
            ):
                objs = []
                expanded_config[argument] = []
                for idx, item in enumerate(config[argument]):
                    obj, obj_config = registry.__getattribute__(
                        f"get_{args[0].type_name()}"
                    )(item, f"{prefix}{argument}.{idx}")
                    objs.append(obj)
                    expanded_config[argument].append(obj_config)
                config[argument] = objs
            elif (
                origin == dict
                and len(args) == 2
                and isinstance(args[1], type)
                and issubclass(args[1], Registrable)
                and isinstance(config[argument], Mapping)
            ):
                objs = {}
                expanded_config[argument] = {}
                for key, val in config[argument].items():
                    obj, obj_config = registry.__getattribute__(
                        f"get_{args[1].type_name()}"
                    )(val, f"{prefix}{argument}.{key}")
                    objs[key] = obj
                    expanded_config[argument][key] = obj_config
                config[argument] = objs

    return config, expanded_config


def get_callable_parsed_args(callable, prefix=None):
    """Helper function that extracts the command line arguments for a given function.

    Args:
        callable (callable): function whose arguments will be inspected to extract
            arguments from the command line.
        prefix (str): Prefix that is attached to the argument names when looking for
            command line arguments.
    """
    signature = inspect.signature(callable)
    arguments = {
        argument: signature.parameters[argument]
        for argument in signature.parameters
        if argument != "self"
    }
    return get_parsed_args(arguments, prefix)


def get_parsed_args(arguments, prefix=None):
    """Helper function that takes a dictionary mapping argument names to types, and
    extracts command line arguments for those arguments. If the dictionary contains
    a key-value pair "bar": int, and the prefix passed is "foo", this function will
    look for a command line argument "\-\-foo.bar". If present, it will cast it to an
    int.

    If the type for a given argument is not one of `int`, `float`, `str`, or `bool`,
    it simply loads the string into python using a yaml loader.

    Args:
        arguments (dict[str, type]): dictionary mapping argument names to types
        prefix (str): prefix that is attached to each argument name before searching
            for command line arguments.
    """
    prefix = "" if prefix is None else f"{prefix}."
    parser = argparse.ArgumentParser()
    for argument in arguments:
        parser.add_argument(f"--{prefix}{argument}")
    parsed_args, _ = parser.parse_known_args()
    parsed_args = vars(parsed_args)

    # Strip the prefix from the parsed arguments and remove arguments not present
    parsed_args = {
        (key[len(prefix) :] if key.startswith(prefix) else key): parsed_args[key]
        for key in parsed_args
        if parsed_args[key] is not None
    }

    for argument in parsed_args:
        expected_type = arguments[argument]
        if isinstance(expected_type, inspect.Parameter):
            expected_type = expected_type.annotation
        if expected_type in [int, str, float]:
            parsed_args[argument] = expected_type(parsed_args[argument])
        elif expected_type is bool:
            value = str(parsed_args[argument]).lower()
            parsed_args[argument] = not ("false".startswith(value) or value == "0")
        else:
            parsed_args[argument] = yaml.safe_load(parsed_args[argument])

    return parsed_args


registry = Registry()



from functools import partial
from typing import List, Tuple, Union

import numpy as np
import torch
from torch import nn

from hive.agents.qnets.noisy_linear import NoisyLinear


class MLPNetwork(nn.Module):
    """Basic MLP neural network architecture.

    Contains a series of :py:class:`torch.nn.Linear` or
    :py:class:`~hive.agents.qnets.noisy_linear.NoisyLinear` layers, each of which
    is followed by a ReLU.
    """

    def __init__(
        self,
        in_dim: Tuple[int],
        hidden_units: Union[int, List[int]] = 256,
        noisy: bool = False,
        std_init: float = 0.5,
    ):
        """
        Args:
            in_dim (tuple[int]): The shape of input observations.
            hidden_units (int | list[int]): The number of neurons for each mlp layer.
            noisy (bool): Whether the MLP should use
                :py:class:`~hive.agents.qnets.noisy_linear.NoisyLinear` layers or normal
                :py:class:`torch.nn.Linear` layers.
            std_init (float): The range for the initialization of the standard deviation of the
                weights in :py:class:`~hive.agents.qnets.noisy_linear.NoisyLinear`.
        """
        super().__init__()
        if isinstance(hidden_units, int):
            hidden_units = [hidden_units]
        linear_fn = partial(NoisyLinear, std_init=std_init) if noisy else nn.Linear
        modules = [linear_fn(np.prod(in_dim), hidden_units[0]), torch.nn.ReLU()]
        for i in range(len(hidden_units) - 1):
            modules.append(linear_fn(hidden_units[i], hidden_units[i + 1]))
            modules.append(torch.nn.ReLU())
        self.network = torch.nn.Sequential(*modules)

    def forward(self, x):
        x = x.float()
        x = torch.flatten(x, start_dim=1)
        return self.network(x)



import os
from collections import deque

import numpy as np
import torch
import yaml

from hive.utils.utils import PACKAGE_ROOT


def load_config(
    config=None,
    preset_config=None,
    agent_config=None,
    env_config=None,
    logger_config=None,
):
    """Used to load config for experiments. Agents, environment, and loggers components
    in main config file can be overrided based on other log files.

    Args:
        config (str): Path to configuration file. Either this or :obj:`preset_config`
            must be passed.
        preset_config (str): Path to a preset hive config. This path should be relative
            to :obj:`hive/configs`. For example, the Atari DQN config would be
            :obj:`atari/dqn.yml`.
        agent_config (str): Path to agent configuration file. Overrides settings in
            base config.
        env_config (str): Path to environment configuration file. Overrides settings in
            base config.
        logger_config (str): Path to logger configuration file. Overrides settings in
            base config.
    """
    if config is not None:
        with open(config) as f:
            yaml_config = yaml.safe_load(f)
    else:
        with open(os.path.join(PACKAGE_ROOT, "configs", preset_config)) as f:
            yaml_config = yaml.safe_load(f)
    if agent_config is not None:
        with open(agent_config) as f:
            if "agents" in yaml_config:
                yaml_config["agents"] = yaml.safe_load(f)
            else:
                yaml_config["agent"] = yaml.safe_load(f)
    if env_config is not None:
        with open(env_config) as f:
            yaml_config["environment"] = yaml.safe_load(f)
    if logger_config is not None:
        with open(logger_config) as f:
            yaml_config["loggers"] = yaml.safe_load(f)
    return yaml_config


class Metrics:
    """Class used to keep track of separate metrics for each agent as well general
    episode metrics.
    """

    def __init__(self, agents, agent_metrics, episode_metrics):
        """
        Args:
            agents (list[Agent]): List of agents for which object will track metrics.
            agent_metrics (list[(str, (callable | obj))]): List of metrics to track
                for each agent. Should be a list of tuples (metric_name, metric_init)
                where metric_init is either the initial value of the metric or a
                callable that takes no arguments and creates the initial metric.
            episode_metrics (list[(str, (callable | obj))]): List of non agent specific
                metrics to keep track of. Should be a list of tuples
                (metric_name, metric_init) where metric_init is either the initial
                value of the metric or a callable with no arguments that creates the
                initial metric.
        """
        self._metrics = {}
        self._agent_metrics = agent_metrics
        self._episode_metrics = episode_metrics
        self._agent_ids = [agent.id for agent in agents]
        self.reset_metrics()

    def reset_metrics(self):
        """Resets all metrics to their initial values."""
        for agent_id in self._agent_ids:
            self._metrics[agent_id] = {}
            for metric_name, metric_value in self._agent_metrics:
                self._metrics[agent_id][metric_name] = (
                    metric_value() if callable(metric_value) else metric_value
                )

        for metric_name, metric_value in self._episode_metrics:
            self._metrics[metric_name] = (
                metric_value() if callable(metric_value) else metric_value
            )

    def get_flat_dict(self):
        """Get a flat dictionary version of the metrics. Each agent metric will be
        prefixed by the agent id.
        """
        metrics = {}
        for metric, _ in self._episode_metrics:
            metrics[metric] = self._metrics[metric]
        for agent_id in self._agent_ids:
            for metric, _ in self._agent_metrics:
                metrics[f"{agent_id}_{metric}"] = self._metrics[agent_id][metric]
        return metrics

    def __getitem__(self, key):
        return self._metrics[key]

    def __setitem__(self, key, value):
        self._metrics[key] = value

    def __repr__(self) -> str:
        return str(self._metrics)


class TransitionInfo:
    """Used to keep track of the most recent transition for each agent.

    Any info that the agent needs to remember for updating can be stored here. Should
    be completely reset between episodes. After any info is extracted, it is
    automatically removed from the object. Also keeps track of which agents have
    started their episodes.

    This object also handles padding and stacking observations for agents.
    """

    def __init__(self, agents, stack_size):
        """
        Args:
            agents (list[Agent]): list of agents that will be kept track of.
            stack_size (int): How many observations will be stacked.
        """
        self._agent_ids = [agent.id for agent in agents]
        self._num_agents = len(agents)
        self._stack_size = stack_size
        self.reset()

    def reset(self):
        """Reset the object by clearing all info."""
        self._transitions = {agent_id: {"reward": 0.0} for agent_id in self._agent_ids}
        self._started = {agent_id: False for agent_id in self._agent_ids}
        self._previous_observations = {
            agent_id: deque(maxlen=self._stack_size - 1) for agent_id in self._agent_ids
        }

    def is_started(self, agent):
        """Check if agent has started its episode.

        Args:
            agent (Agent): Agent to check.
        """
        return self._started[agent.id]

    def start_agent(self, agent):
        """Set the agent's start flag to true.

        Args:
            agent (Agent): Agent to start.
        """
        self._started[agent.id] = True

    def record_info(self, agent, info):
        """Update some information for the agent.

        Args:
            agent (Agent): Agent to update.
            info (dict): Info to add to the agent's state.
        """
        self._transitions[agent.id].update(info)
        if "observation" in info:
            self._previous_observations[agent.id].append(info["observation"])

    def update_reward(self, agent, reward):
        """Add a reward to the agent.

        Args:
            agent (Agent): Agent to update.
            reward (float): Reward to add to agent.
        """
        self._transitions[agent.id]["reward"] += reward

    def update_all_rewards(self, rewards):
        """Update the rewards for all agents. If rewards is list, it updates the rewards
        according to the order of agents provided in the initializer. If rewards is a
        dict, the keys should be the agent ids for the agents and the values should be
        the rewards for those agents. If rewards is a float or int, every agent is
        updated with that reward.

        Args:
            rewards (float | list | np.ndarray | dict): Rewards to update agents with.
        """
        if isinstance(rewards, list) or isinstance(rewards, np.ndarray):
            for idx, agent_id in enumerate(self._agent_ids):
                self._transitions[agent_id]["reward"] += rewards[idx]
        elif isinstance(rewards, int) or isinstance(rewards, float):
            for agent_id in self._agent_ids:
                self._transitions[agent_id]["reward"] += rewards
        else:
            for agent_id in rewards:
                self._transitions[agent_id]["reward"] += rewards[agent_id]

    def get_info(self, agent, done=False):
        """Get all the info for the agent, and reset the info for that agent. Also adds
        a done value to the info dictionary that is based on the done parameter to the
        function.

        Args:
            agent (Agent): Agent to get transition update info for.
            done (bool): Whether this transition is terminal.
        """
        info = self._transitions[agent.id]
        info["done"] = done
        self._transitions[agent.id] = {"reward": 0.0}
        return info

    def get_stacked_state(self, agent, observation):
        """Create a stacked state for the agent. The previous observations recorded
        by this agent are stacked with the current observation. If not enough
        observations have been recorded, zero arrays are appended.

        Args:
            agent (Agent): Agent to get stacked state for.
            observation: Current observation.
        """

        if self._stack_size == 1:
            return observation
        while len(self._previous_observations[agent.id]) < self._stack_size - 1:
            self._previous_observations[agent.id].append(zeros_like(observation))

        stacked_observation = concatenate(
            list(self._previous_observations[agent.id]) + [observation]
        )
        return stacked_observation

    def __repr__(self):
        return str(
            {
                "transtions": self._transitions,
                "started": self._started,
                "previous_observations": self._previous_observations,
            }
        )


def zeros_like(x):
    """Create a zero state like some state. This handles slightly more complex
    objects such as lists and dictionaries of numpy arrays and torch Tensors.

    Args:
        x (np.ndarray | torch.Tensor | dict | list): State used to define
            structure/state of zero state.
    """
    if isinstance(x, np.ndarray):
        return np.zeros_like(x)
    elif isinstance(x, torch.Tensor):
        return torch.zeros_like(x)
    elif isinstance(x, dict):
        return {k: zeros_like(v) for k, v in x.items()}
    elif isinstance(x, list):
        return [zeros_like(item) for item in x]
    else:
        return 0


def concatenate(xs):
    """Concatenates numpy arrays or dictionaries of numpy arrays.

    Args:
        xs (list): List of objects to concatenate.
    """

    if len(xs) == 0:
        return np.array([])

    if isinstance(xs[0], dict):
        return {k: np.concatenate([x[k] for x in xs], axis=0) for k in xs[0]}
    else:
        return np.concatenate(xs, axis=0)



import os
from typing import Dict, Tuple

import numpy as np

from hive.replays.circular_replay import CircularReplayBuffer
from hive.utils.torch_utils import numpify


class PrioritizedReplayBuffer(CircularReplayBuffer):
    """Implements a replay with prioritized sampling. See
    https://arxiv.org/abs/1511.05952
    """

    def __init__(
        self,
        capacity: int,
        beta: float = 0.5,
        stack_size: int = 1,
        n_step: int = 1,
        gamma: float = 0.9,
        observation_shape: Tuple = (),
        observation_dtype: type = np.uint8,
        action_shape: Tuple = (),
        action_dtype: type = np.int8,
        reward_shape: Tuple = (),
        reward_dtype: type = np.float32,
        extra_storage_types: Dict = None,
        num_players_sharing_buffer=None,
    ):
        """
        Args:
            capacity (int): Total number of observations that can be stored in the
                buffer. Note, this is not the same as the number of transitions that
                can be stored in the buffer.
            beta (float): Parameter controlling level of prioritization.
            stack_size (int): The number of frames to stack to create an observation.
            n_step (int): Horizon used to compute n-step return reward
            gamma (float): Discounting factor used to compute n-step return reward
            observation_shape: Shape of observations that will be stored in the buffer.
            observation_dtype: Type of observations that will be stored in the buffer.
                This can either be the type itself or string representation of the
                type. The type can be either a native python type or a numpy type. If
                a numpy type, a string of the form np.uint8 or numpy.uint8 is
                acceptable.
            action_shape: Shape of actions that will be stored in the buffer.
            action_dtype: Type of actions that will be stored in the buffer. Format is
                described in the description of observation_dtype.
            action_shape: Shape of actions that will be stored in the buffer.
            action_dtype: Type of actions that will be stored in the buffer. Format is
                described in the description of observation_dtype.
            reward_shape: Shape of rewards that will be stored in the buffer.
            reward_dtype: Type of rewards that will be stored in the buffer. Format is
                described in the description of observation_dtype.
            extra_storage_types (dict): A dictionary describing extra items to store
                in the buffer. The mapping should be from the name of the item to a
                (type, shape) tuple.
            num_players_sharing_buffer (int): Number of agents that share their
                buffers. It is used for self-play.
        """
        super().__init__(
            capacity=capacity,
            stack_size=stack_size,
            n_step=n_step,
            gamma=gamma,
            observation_shape=observation_shape,
            observation_dtype=observation_dtype,
            action_shape=action_shape,
            action_dtype=action_dtype,
            reward_shape=reward_shape,
            reward_dtype=reward_dtype,
            extra_storage_types=extra_storage_types,
            num_players_sharing_buffer=num_players_sharing_buffer,
        )
        self._sum_tree = SumTree(self._capacity)
        self._beta = beta

    def set_beta(self, beta):
        self._beta = beta

    def _add_transition(self, priority=None, **transition):
        if priority is None:
            priority = self._sum_tree.max_recorded_priority
        self._sum_tree.set_priority(self._cursor, priority)
        super()._add_transition(**transition)

    def _pad_buffer(self, pad_length):
        for _ in range(pad_length):
            transition = {
                key: np.zeros_like(self._storage[key][0]) for key in self._storage
            }
            transition["priority"] = 0
            self._add_transition(**transition)

    def _sample_indices(self, batch_size):
        indices = self._sum_tree.stratified_sample(batch_size)
        indices = self._filter_transitions(indices)
        while len(indices) < batch_size:
            new_indices = self._sum_tree.sample(batch_size - len(indices))
            new_indices = self._filter_transitions(new_indices)
            indices = np.concatenate([indices, new_indices])
        return indices

    def _filter_transitions(self, indices):
        indices = super()._filter_transitions(indices - (self._stack_size - 1)) + (
            self._stack_size - 1
        )
        if self._num_added < self._capacity:
            indices = indices[indices < self._cursor - self._n_step]
            indices = indices[indices >= self._stack_size - 1]
        else:
            low = (self._cursor - self._n_step) % self._capacity
            high = (self._cursor + self._stack_size - 1) % self._capacity
            if low < high:
                indices = indices[np.logical_or(indices < low, indices > high)]
            else:
                indices = indices[~np.logical_or(indices >= low, indices <= high)]

        return indices

    def sample(self, batch_size):
        batch = super().sample(batch_size)
        indices = batch["indices"]
        priorities = self._sum_tree.get_priorities(indices)
        weights = (1.0 / (priorities + 1e-10)) ** self._beta
        weights /= np.max(weights)
        batch["weights"] = weights
        return batch

    def update_priorities(self, indices, priorities):
        """Update the priorities of the transitions at the specified indices.

        Args:
            indices: Which transitions to update priorities for. Can be numpy array
                or torch tensor.
            priorities: What the priorities should be updated to. Can be numpy array
                or torch tensor.
        """
        indices = numpify(indices)
        priorities = numpify(priorities)
        indices, unique_idxs = np.unique(indices, return_index=True)
        priorities = priorities[unique_idxs]
        self._sum_tree.set_priority(indices, priorities)

    def save(self, dname):
        super().save(dname)
        self._sum_tree.save(dname)

    def load(self, dname):
        super().load(dname)
        self._sum_tree.load(dname)


class SumTree:
    """Data structure used to implement prioritized sampling. It is implemented
    as a tree where the value of each node is the sum of the values of the subtree
    of the node.
    """

    def __init__(self, capacity: int):
        self._capacity = capacity
        self._depth = int(np.ceil(np.log2(capacity))) + 1
        self._tree = np.zeros(2 ** self._depth - 1)
        self._last_level_start = 2 ** (self._depth - 1) - 1
        self._priorities = self._tree[
            self._last_level_start : self._last_level_start + self._capacity
        ]
        self.max_recorded_priority = 1.0

    def set_priority(self, indices, priorities):
        """Sets the priorities for the given indices.

        Args:
            indices (np.ndarray): Which transitions to update priorities for.
            priorities (np.ndarray): What the priorities should be updated to.
        """
        self.max_recorded_priority = max(self.max_recorded_priority, np.max(priorities))
        indices = self._last_level_start + indices
        diffs = priorities - self._tree[indices]
        for _ in range(self._depth):
            np.add.at(self._tree, indices, diffs)
            indices = (indices - 1) // 2

    def sample(self, batch_size):
        """Sample elements from the sum tree with probability proportional to their
        priority.

        Args:
            batch_size (int): The number of elements to sample.
        """
        indices = self.extract(np.random.rand(batch_size))
        return indices

    def stratified_sample(self, batch_size):
        """Performs stratified sampling using the sum tree.

        Args:
            batch_size (int): The number of elements to sample.
        """
        query_values = (np.arange(batch_size) + np.random.rand(batch_size)) / batch_size
        indices = self.extract(query_values)
        return indices

    def extract(self, queries):
        """Get the elements in the sum tree that correspond to the query.
        For each query, the element that is selected is the one with the greatest
        sum of "previous" elements in the tree, but also such that the sum is not
        a greater proportion of the total sum of priorities than the query.

        Args:
            queries (np.ndarray): Queries to extract. Each element should be
                between 0 and 1.
        """
        queries *= self._tree[0]
        indices = np.zeros(queries.shape[0], dtype=np.int64)
        for i in range(self._depth - 1):
            indices = indices * 2 + 1
            left_child_values = self._tree[indices]
            branch_right = (queries > left_child_values).nonzero()
            indices[branch_right] += 1
            queries[branch_right] -= left_child_values[branch_right]
        return indices - self._last_level_start

    def get_priorities(self, indices):
        """Get the priorities of the elements at indicies.

        Args:
            indices (np.ndarray): The indices to query.
        """
        return self._priorities[indices]

    def save(self, dname):
        np.save(os.path.join(dname, "sumtree.npy"), self._tree)

    def load(self, dname):
        self._tree = np.load(os.path.join(dname, "sumtree.npy"))
        self._priorities = self._tree[
            self._last_level_start : self._last_level_start + self._capacity
        ]



from importlib import import_module

import numpy as np

from hive.envs import BaseEnv
from hive.envs.env_spec import EnvSpec


class PettingZooEnv(BaseEnv):
    """
    PettingZoo environment from https://github.com/PettingZoo-Team/PettingZoo

    For now, we only support environments from PettingZoo with discrete actions.
    """

    def __init__(
        self,
        env_name,
        env_family,
        num_players,
        **kwargs,
    ):
        """
        Args:
            env_name (str): Name of the environment
            env_family (str): Family of the environment such as "Atari",
            "Classic", "SISL", "Butterfly", "MAgent", and "MPE".
            num_players (int): Number of learning agents
        """
        self._env_family = env_family
        self.create_env(env_name, num_players, **kwargs)
        self._env_spec = self.create_env_spec(env_name, **kwargs)
        super().__init__(self.create_env_spec(env_name, **kwargs), num_players)

    def create_env(self, env_name, num_players, **kwargs):
        env_module = import_module("pettingzoo." + self._env_family + "." + env_name)
        self._env = env_module.env(players=num_players)

    def create_env_spec(self, env_name, **kwargs):
        """
        Each family of environments have their own type of observations and actions.
        You can add support for more families here by modifying obs_dim and act_dim.
        """
        if self._env_family in ["classic"]:
            obs_dim = [
                space["observation"].shape
                for space in self._env.observation_spaces.values()
            ]
        elif self._env_family in ["sisl"]:
            obs_dim = [space.shape for space in self._env.observation_spaces.values()]
        else:
            raise ValueError(
                f"Hive does not support {self._env_family} environments from PettingZoo yet."
            )
        act_dim = [space.n for space in self._env.action_spaces.values()]
        return EnvSpec(
            env_name=env_name,
            obs_dim=obs_dim,
            act_dim=act_dim,
        )

    def reset(self):
        self._env.reset()
        observation, _, _, _ = self._env.last()
        for key in observation.keys():
            observation[key] = np.array(observation[key], dtype=np.uint8)
        self._turn = self._env.agents.index(self._env.agent_selection)

        return observation, self._turn

    def step(self, action):
        self._env.step(action)
        observation, _, done, info = self._env.last()
        self._turn = (self._turn + 1) % self._num_players
        for key in observation.keys():
            observation[key] = np.array(observation[key], dtype=np.uint8)
        return (
            observation,
            [self._env.rewards[agent] for agent in self._env.agents],
            done,
            self._turn,
            info,
        )

    def render(self, mode="rgb_array"):
        return self._env.render(mode=mode)

    def seed(self, seed=None):
        self._env.seed(seed=seed)

    def close(self):
        self._env.close()



import os

import setuptools

with open("version.txt") as f:
    VERSION = f.read().strip()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

atari = ["gym[atari,accept-rom-license]~=0.21", "opencv-python~=4.0"]
gym_minigrid = ["gym-minigrid~=1.0"]
petting_zoo = ["pettingzoo[sisl,atari,classic]~=1.11"]
test = ["pytest~=6.2", "pytest-lazy-fixture~=0.6"]


setuptools.setup(
    name="rlhive",
    version=VERSION,
    description="A package to support RL research",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chandar-lab/RLHive",
    project_urls={
        "Bug Tracker": "https://github.com/chandar-lab/RLHive/issues",
    },
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={"hive": ["configs/**.yml"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.6",
    install_requires=[
        "gym~=0.21",
        "numpy~=1.18",
        "PyYAML~=5.4",
        "torch~=1.6",
        "wandb>=0.10.30",
        "matplotlib~=3.0",
        "pandas~=1.0",
    ],
    extras_require={
        "atari": atari,
        "gym_minigrid": gym_minigrid,
        "petting_zoo": petting_zoo,
        "test": test,
        "all": atari + gym_minigrid + petting_zoo + test,
    },
    entry_points={
        "console_scripts": [
            "hive_single_agent_loop = hive.runners.single_agent_loop:main",
            "hive_multi_agent_loop = hive.runners.multi_agent_loop:main",
        ]
    },
)



import numpy as np
import torch

from hive.agents.rainbow import RainbowDQNAgent


class LegalMovesRainbowAgent(RainbowDQNAgent):
    """A Rainbow agent which supports games with legal actions."""

    def create_q_networks(self, representation_net):
        """Creates the qnet and target qnet."""
        super().create_q_networks(representation_net)
        self._qnet = LegalMovesHead(self._qnet)
        self._target_qnet = LegalMovesHead(self._target_qnet)

    def preprocess_update_info(self, update_info):
        preprocessed_update_info = {
            "observation": update_info["observation"]["observation"],
            "action": update_info["action"],
            "reward": update_info["reward"],
            "done": update_info["done"],
            "action_mask": action_encoding(update_info["observation"]["action_mask"]),
        }
        if "agent_id" in update_info:
            preprocessed_update_info["agent_id"] = int(update_info["agent_id"])
        return preprocessed_update_info

    def preprocess_update_batch(self, batch):
        for key in batch:
            batch[key] = torch.tensor(batch[key], device=self._device)
        return (
            (batch["observation"], batch["action_mask"]),
            (batch["next_observation"], batch["next_action_mask"]),
            batch,
        )

    @torch.no_grad()
    def act(self, observation):
        if self._training:
            if not self._learn_schedule.get_value():
                epsilon = 1.0
            elif not self._use_eps_greedy:
                epsilon = 0.0
            else:
                epsilon = self._epsilon_schedule.update()
            if self._logger.update_step(self._timescale):
                self._logger.log_scalar("epsilon", epsilon, self._timescale)
        else:
            epsilon = self._test_epsilon

        vectorized_observation = torch.tensor(
            np.expand_dims(observation["observation"], axis=0), device=self._device
        ).float()
        legal_moves_as_int = [
            i for i, x in enumerate(observation["action_mask"]) if x == 1
        ]
        encoded_legal_moves = torch.tensor(
            action_encoding(observation["action_mask"]), device=self._device
        ).float()
        qvals = self._qnet(vectorized_observation, encoded_legal_moves).cpu()

        if self._rng.random() < epsilon:
            action = np.random.choice(legal_moves_as_int).item()
        else:
            action = torch.argmax(qvals).item()

        if (
            self._training
            and self._logger.should_log(self._timescale)
            and self._state["episode_start"]
        ):
            self._logger.log_scalar("train_qval", torch.max(qvals), self._timescale)
            self._state["episode_start"] = False

        return action


class LegalMovesHead(torch.nn.Module):
    def __init__(self, base_network):
        super().__init__()
        self.base_network = base_network

    def forward(self, x, legal_moves):
        x = self.base_network(x)
        return x + legal_moves

    def dist(self, x, legal_moves):
        return self.base_network.dist(x)


def action_encoding(action_mask):
    encoded_action_mask = np.zeros(action_mask.shape)
    encoded_action_mask[action_mask == 0] = -np.inf
    return encoded_action_mask



from hive.envs import GymEnv, ParallelEnv
from hive.envs.wrappers.gym_wrappers import FlattenWrapper, PermuteImageWrapper
from marlgrid import envs


class MarlGridEnv(ParallelEnv, GymEnv):
    """MarlGrid environment from https://github.com/kandouss/marlgrid/.

    The environment can either be initialized with the name of a preregistered
    environment from
    https://github.com/kandouss/marlgrid/blob/master/marlgrid/envs/__init__.py,
    or can be created using a config. See the original repo for details.
    """

    def create_env(self, env_name, randomize_seed=True, flatten=False, **kwargs):
        """
        Args:
            env_name: The name of the environment.
            randomize_seed: Whether to use a random random seed for the environment.
            flatten: Whether to flatten the observations.
        """
        if env_name is None:
            self._env = envs.env_from_config(kwargs, randomize_seed=randomize_seed)
        else:
            super().create_env(env_name, **kwargs)

        self._env = PermuteImageWrapper(self._env)
        if flatten:
            self._env = FlattenWrapper(self._env)

    def create_env_spec(self, name, **kwargs):
        return super().create_env_spec(
            name if name is not None else f"Marlgrid_{str(kwargs)}", **kwargs
        )



import argparse
import copy

from hive import agents as agent_lib
from hive import envs
from hive.runners.base import Runner
from hive.runners.utils import TransitionInfo, load_config
from hive.utils import experiment, loggers, schedule, utils
from hive.utils.registry import get_parsed_args


class SingleAgentRunner(Runner):
    """Runner class used to implement a sinle-agent training loop."""

    def __init__(
        self,
        environment,
        agent,
        logger,
        experiment_manager,
        train_steps,
        test_frequency,
        test_episodes,
        stack_size,
        max_steps_per_episode=27000,
    ):
        """Initializes the Runner object.

        Args:
            environment (BaseEnv): Environment used in the training loop.
            agent (Agent): Agent that will interact with the environment
            logger (ScheduledLogger): Logger object used to log metrics.
            experiment_manager (Experiment): Experiment object that saves the state of
                the training.
            train_steps (int): How many steps to train for. If this is -1, there is no
                limit for the number of training steps.
            test_frequency (int): After how many training steps to run testing
                episodes. If this is -1, testing is not run.
            test_episodes (int): How many episodes to run testing for duing each test
                phase.
            stack_size (int): The number of frames in an observation sent to an agent.
            max_steps_per_episode (int): The maximum number of steps to run an episode
                for.
        """
        super().__init__(
            environment,
            agent,
            logger,
            experiment_manager,
            train_steps,
            test_frequency,
            test_episodes,
            max_steps_per_episode,
        )
        self._transition_info = TransitionInfo(self._agents, stack_size)

    def run_one_step(self, observation, episode_metrics):
        """Run one step of the training loop.

        Args:
            observation: Current observation that the agent should create an action
                for.
            episode_metrics (Metrics): Keeps track of metrics for current episode.
        """
        super().run_one_step(observation, 0, episode_metrics)
        agent = self._agents[0]
        stacked_observation = self._transition_info.get_stacked_state(
            agent, observation
        )
        action = agent.act(stacked_observation)
        next_observation, reward, done, _, other_info = self._environment.step(action)

        info = {
            "observation": observation,
            "reward": reward,
            "action": action,
            "done": done,
            "info": other_info,
        }
        if self._training:
            agent.update(copy.deepcopy(info))

        self._transition_info.record_info(agent, info)
        episode_metrics[agent.id]["reward"] += info["reward"]
        episode_metrics[agent.id]["episode_length"] += 1
        episode_metrics["full_episode_length"] += 1

        return done, next_observation

    def run_episode(self):
        """Run a single episode of the environment."""
        episode_metrics = self.create_episode_metrics()
        done = False
        observation, _ = self._environment.reset()
        self._transition_info.reset()
        self._transition_info.start_agent(self._agents[0])
        steps = 0
        # Run the loop until the episode ends or times out
        while not done and steps < self._max_steps_per_episode:
            done, observation = self.run_one_step(observation, episode_metrics)
            steps += 1

        return episode_metrics


def set_up_experiment(config):
    """Returns a :py:class:`SingleAgentRunner` object based on the config and any
    command line arguments.

    Args:
        config: Configuration for experiment.
    """

    args = get_parsed_args(
        {
            "seed": int,
            "train_steps": int,
            "test_frequency": int,
            "test_episodes": int,
            "max_steps_per_episode": int,
            "stack_size": int,
            "resume": bool,
            "run_name": str,
            "save_dir": str,
        }
    )
    config.update(args)
    full_config = utils.Chomp(copy.deepcopy(config))
    if "seed" in config:
        utils.seeder.set_global_seed(config["seed"])

    environment, full_config["environment"] = envs.get_env(
        config["environment"], "environment"
    )
    env_spec = environment.env_spec

    # Set up loggers
    logger_config = config.get("loggers", {"name": "NullLogger"})
    if logger_config is None or len(logger_config) == 0:
        logger_config = {"name": "NullLogger"}
    if isinstance(logger_config, list):
        logger_config = {
            "name": "CompositeLogger",
            "kwargs": {"logger_list": logger_config},
        }

    logger, full_config["loggers"] = loggers.get_logger(logger_config, "loggers")

    # Set up agent
    if config.get("stack_size", 1) > 1:
        config["agent"]["kwargs"]["obs_dim"] = (
            config["stack_size"] * env_spec.obs_dim[0][0],
            *env_spec.obs_dim[0][1:],
        )
    else:
        config["agent"]["kwargs"]["obs_dim"] = env_spec.obs_dim[0]
    config["agent"]["kwargs"]["act_dim"] = env_spec.act_dim[0]
    config["agent"]["kwargs"]["logger"] = logger
    if "replay_buffer" in config["agent"]["kwargs"]:
        replay_args = config["agent"]["kwargs"]["replay_buffer"]["kwargs"]
        replay_args["observation_shape"] = env_spec.obs_dim[0]
    agent, full_config["agent"] = agent_lib.get_agent(config["agent"], "agent")

    # Set up experiment manager
    saving_schedule, full_config["saving_schedule"] = schedule.get_schedule(
        config["saving_schedule"], "saving_schedule"
    )
    experiment_manager = experiment.Experiment(
        config["run_name"], config["save_dir"], saving_schedule
    )
    experiment_manager.register_experiment(
        config=full_config,
        logger=logger,
        agents=agent,
    )
    # Set up runner
    runner = SingleAgentRunner(
        environment,
        agent,
        logger,
        experiment_manager,
        config.get("train_steps", -1),
        config.get("test_frequency", -1),
        config.get("test_episodes", 1),
        config.get("stack_size", 1),
        config.get("max_steps_per_episode", 1e9),
    )
    if config.get("resume", False):
        runner.resume()

    return runner


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config")
    parser.add_argument("-p", "--preset-config")
    parser.add_argument("-a", "--agent-config")
    parser.add_argument("-e", "--env-config")
    parser.add_argument("-l", "--logger-config")
    args, _ = parser.parse_known_args()
    if args.config is None and args.preset_config is None:
        raise ValueError("Config needs to be provided")
    config = load_config(
        args.config,
        args.preset_config,
        args.agent_config,
        args.env_config,
        args.logger_config,
    )
    runner = set_up_experiment(config)
    runner.run_training()


if __name__ == "__main__":
    main()



from hive.utils.registry import registry
from hive.agents.qnets.atari import NatureAtariDQNModel
from hive.agents.qnets.base import FunctionApproximator
from hive.agents.qnets.conv import ConvNetwork
from hive.agents.qnets.mlp import MLPNetwork

registry.register_all(
    FunctionApproximator,
    {
        "MLPNetwork": FunctionApproximator(MLPNetwork),
        "ConvNetwork": FunctionApproximator(ConvNetwork),
        "NatureAtariDQNModel": FunctionApproximator(NatureAtariDQNModel),
    },
)

get_qnet = getattr(registry, f"get_{FunctionApproximator.type_name()}")



import abc

from hive.utils.registry import Registrable


class BaseReplayBuffer(abc.ABC, Registrable):
    """Base class for replay buffers. Every implemented buffer should be a subclass of this class."""

    @abc.abstractmethod
    def add(self, **data):
        """
        Adds data to the buffer

        Args:
            data: data to add to the replay buffer. Subclasses can define this class
                signature based on use case.
        """

    @abc.abstractmethod
    def sample(self, batch_size):
        """
        sample a minibatch

        Args:
            batch_size (int): the number of transitions to sample.
        """

    @abc.abstractmethod
    def size(self):
        """
        Returns the number of transitions stored in the buffer.
        """

    @abc.abstractmethod
    def save(self, dname):
        """
        Saves buffer checkpointing information to file for future loading.

        Args:
            dname (str): directory where agent should save all relevant info.
        """

    @abc.abstractmethod
    def load(self, dname):
        """
        Loads buffer from file.

        Args:
            dname (str): directory name where buffer checkpoint info is stored.

        Returns:
            True if successfully loaded the buffer. False otherwise.
        """

    @classmethod
    def type_name(cls):
        """
        Returns: "replay"
        """
        return "replay"



import abc

from hive.utils.registry import Registrable, registry


class Schedule(abc.ABC, Registrable):
    @abc.abstractmethod
    def get_value(self):
        """Returns the current value of the variable we are tracking"""
        pass

    @abc.abstractmethod
    def update(self):
        """Update the value of the variable we are tracking and return the updated value.
        The first call to update will return the initial value of the schedule."""
        pass

    @classmethod
    def type_name(cls):
        return "schedule"


class LinearSchedule(Schedule):
    """Defines a linear schedule between two values over some number of steps.

    If updated more than the defined number of steps, the schedule stays at the
    end value.
    """

    def __init__(self, init_value, end_value, steps):
        """
        Args:
            init_value (int | float): Starting value for schedule.
            end_value (int | float): End value for schedule.
            steps (int): Number of steps for schedule. Should be positive.
        """
        steps = max(int(steps), 1)
        self._delta = (end_value - init_value) / steps
        self._end_value = end_value
        self._value = init_value - self._delta

    def get_value(self):
        return self._value

    def update(self):
        if self._value == self._end_value:
            return self._value

        self._value += self._delta

        # Check if value is over the end_value
        if ((self._value - self._end_value) > 0) == (self._delta > 0):
            self._value = self._end_value
        return self._value

    def __repr__(self):
        return (
            f"<class {type(self).__name__}"
            f" value={self.get_value()}"
            f" delta={self._delta}"
            f" end_value={self._end_value}>"
        )


class ConstantSchedule(Schedule):
    """Returns a constant value over the course of the schedule"""

    def __init__(self, value):
        """
        Args:
            value: The value to be returned.
        """
        self._value = value

    def get_value(self):
        return self._value

    def update(self):
        return self._value

    def __repr__(self):
        return f"<class {type(self).__name__} value={self.get_value()}>"


class SwitchSchedule(Schedule):
    """Returns one value for the first part of the schedule. After the defined
    number of steps is reached, switches to returning a second value.
    """

    def __init__(self, off_value, on_value, steps):
        """
        Args:
            off_value: The value to be returned in the first part of the schedule.
            on_value: The value to be returned in the second part of the schedule.
            steps (int): The number of steps after which to switch from the off
                value to the on value.
        """

        self._steps = 0
        self._flip_step = steps
        self._off_value = off_value
        self._on_value = on_value

    def get_value(self):
        if self._steps <= self._flip_step:
            return self._off_value
        else:
            return self._on_value

    def update(self):
        self._steps += 1
        value = self.get_value()
        return value

    def __repr__(self):
        return (
            f"<class {type(self).__name__}"
            f" value={self.get_value()}"
            f" steps={self._steps}"
            f" off_value={self._off_value}"
            f" on_value={self._on_value}"
            f" flip_step={self._flip_step}>"
        )


class DoublePeriodicSchedule(Schedule):
    """Returns off value for off period, then switches to returning on value for on
    period. Alternates between the two.
    """

    def __init__(self, off_value, on_value, off_period, on_period):
        """
        Args:
            on_value: The value to be returned for the on period.
            off_value: The value to be returned for the off period.
            on_period (int): the number of steps in the on period.
            off_period (int): the number of steps in the off period.
        """
        self._steps = -1
        self._off_period = off_period
        self._total_period = self._off_period + on_period
        self._off_value = off_value
        self._on_value = on_value

    def get_value(self):
        if (self._steps % self._total_period) < self._off_period:
            return self._off_value
        else:
            return self._on_value

    def update(self):
        self._steps += 1
        return self.get_value()

    def __repr__(self):
        return (
            f"<class {type(self).__name__}"
            f" value={self.get_value()}"
            f" steps={self._steps}"
            f" off_value={self._off_value}"
            f" on_value={self._on_value}"
            f" off_period={self._off_period}"
            f" on_period={self._total_period - self._off_period}>"
        )


class PeriodicSchedule(DoublePeriodicSchedule):
    """Returns one value on the first step of each period of a predefined number of
    steps. Returns another value otherwise.
    """

    def __init__(self, off_value, on_value, period):
        """
        Args:
            on_value: The value to be returned on the first step of each period.
            off_value: The value to be returned for every other step in the period.
            period (int): The number of steps in the period.
        """
        super().__init__(off_value, on_value, period - 1, 1)

    def __repr__(self):
        return (
            f"<class {type(self).__name__}"
            f" value={self.get_value()}"
            f" steps={self._steps}"
            f" off_value={self._off_value}"
            f" on_value={self._on_value}"
            f" period={self._off_period + 1}>"
        )


registry.register_all(
    Schedule,
    {
        "LinearSchedule": LinearSchedule,
        "ConstantSchedule": ConstantSchedule,
        "SwitchSchedule": SwitchSchedule,
        "PeriodicSchedule": PeriodicSchedule,
        "DoublePeriodicSchedule": DoublePeriodicSchedule,
    },
)

get_schedule = getattr(registry, f"get_{Schedule.type_name()}")



import ale_py
import cv2
import numpy as np

from hive.envs.env_spec import EnvSpec
from hive.envs.gym_env import GymEnv


class AtariEnv(GymEnv):
    """
    Class for loading Atari environments.

    Adapted from the Dopamine's Atari preprocessing code:
    https://github.com/google/dopamine/blob/6fbb58ad9bc1340f42897e8a551f85a01fb142ce/dopamine/discrete_domains/atari_lib.py
    Licensed under Apache 2.0, https://github.com/google/dopamine/blob/master/LICENSE
    """

    def __init__(
        self,
        env_name,
        frame_skip=4,
        screen_size=84,
        sticky_actions=True,
    ):
        """
        Args:
            env_name (str): Name of the environment
            sticky_actions (bool): Whether to use sticky_actions as per Machado et al.
            frame_skip (int): Number of times the agent takes the same action in the environment
            screen_size (int): Size of the resized frames from the environment
        """
        env_version = "v0" if sticky_actions else "v4"
        full_env_name = "{}NoFrameskip-{}".format(env_name, env_version)

        if frame_skip <= 0:
            raise ValueError(
                "Frame skip should be strictly positive, got {}".format(frame_skip)
            )
        if screen_size <= 0:
            raise ValueError(
                "Target screen size should be strictly positive, got {}".format(
                    screen_size
                )
            )

        self.frame_skip = frame_skip
        self.screen_size = screen_size

        super().__init__(full_env_name)

    def create_env_spec(self, env_name, **kwargs):
        obs_spaces = self._env.observation_space.shape
        # Used for storing and pooling over two consecutive observations
        self.screen_buffer = [
            np.empty((obs_spaces[0], obs_spaces[1]), dtype=np.uint8),
            np.empty((obs_spaces[0], obs_spaces[1]), dtype=np.uint8),
        ]

        act_spaces = [self._env.action_space]
        return EnvSpec(
            env_name=env_name,
            obs_dim=[(1, self.screen_size, self.screen_size)],
            act_dim=[space.n for space in act_spaces],
        )

    def reset(self):
        self._env.reset()
        self._get_observation_screen(self.screen_buffer[1])
        self.screen_buffer[0].fill(0)
        return self._pool_and_resize(), self._turn

    def step(self, action=None):
        """
        Remarks:
            * Executes the action for :attr:`self.frame_skips` steps in the the
              environment.
            * This may execute fewer than self.frame_skip steps in the environment, if
              the done state is reached.
            * In this case the returned observation should be ignored.
        """
        assert action is not None

        accumulated_reward = 0.0
        done = False
        info = {}

        for time_step in range(self.frame_skip):
            _, reward, done, info = self._env.step(action)
            accumulated_reward += reward

            if done:
                break
            elif time_step >= self.frame_skip - 2:
                t = time_step - (self.frame_skip - 2)
                self._get_observation_screen(self.screen_buffer[t])

        observation = self._pool_and_resize()

        return observation, accumulated_reward, done, self._turn, info

    def _get_observation_screen(self, output):
        """Get the screen input of the current observation given empty numpy array in grayscale.

        Args:
          output (np.ndarray): screen buffer to hold the returned observation.

        Returns:
          observation (np.ndarray): the current observation in grayscale.
        """
        self._env.ale.getScreenGrayscale(output)
        return output

    def _pool_and_resize(self):
        """Transforms two frames into a Nature DQN observation.

        Returns:
          transformed_screen (np.ndarray): pooled, resized screen.
        """
        # Pool if there are enough screens to do so.
        if self.frame_skip > 1:
            np.maximum(
                self.screen_buffer[0], self.screen_buffer[1], out=self.screen_buffer[1]
            )

        transformed_image = cv2.resize(
            self.screen_buffer[1],
            (self.screen_size, self.screen_size),
            interpolation=cv2.INTER_AREA,
        )
        int_image = np.asarray(transformed_image, dtype=np.uint8)
        return np.expand_dims(int_image, axis=0)



from abc import ABC
from typing import List
from hive.agents.agent import Agent
from hive.envs.base import BaseEnv

from hive.runners.utils import Metrics
from hive.utils import schedule
from hive.utils.experiment import Experiment
from hive.utils.loggers import ScheduledLogger


class Runner(ABC):
    """Base Runner class used to implement a training loop.

    Different types of training loops can be created by overriding the relevant
    functions.
    """

    def __init__(
        self,
        environment: BaseEnv,
        agents: List[Agent],
        logger: ScheduledLogger,
        experiment_manager: Experiment,
        train_steps: int = 1000000,
        test_frequency: int = 10000,
        test_episodes: int = 1,
        max_steps_per_episode: int = 27000,
    ):
        """
        Args:
            environment (BaseEnv): Environment used in the training loop.
            agents (list[Agent]): List of agents that interact with the environment.
            logger (ScheduledLogger): Logger object used to log metrics.
            experiment_manager (Experiment): Experiment object that saves the state of
                the training.
            train_steps (int): How many steps to train for. If this is -1, there is no
                limit for the number of training steps.
            test_frequency (int): After how many training steps to run testing episodes.
                If this is -1, testing is not run.
            test_episodes (int): How many episodes to run testing for.
        """
        self._environment = environment
        if isinstance(agents, list):
            self._agents = agents
        else:
            self._agents = [agents]
        self._logger = logger
        self._experiment_manager = experiment_manager
        if train_steps == -1:
            self._train_schedule = schedule.ConstantSchedule(True)
        else:
            self._train_schedule = schedule.SwitchSchedule(True, False, train_steps)
        if test_frequency == -1:
            self._test_schedule = schedule.ConstantSchedule(False)
        else:
            self._test_schedule = schedule.PeriodicSchedule(False, True, test_frequency)
        self._test_episodes = test_episodes
        self._max_steps_per_episode = max_steps_per_episode

        self._experiment_manager.experiment_state.update(
            {
                "train_schedule": self._train_schedule,
                "test_schedule": self._test_schedule,
            }
        )
        self._logger.register_timescale("train")
        self._logger.register_timescale("test")
        self._training = True
        self._save_experiment = False
        self._run_testing = False

    def train_mode(self, training):
        """If training is true, sets all agents to training mode. If training is false,
        sets all agents to eval mode.

        Args:
            training (bool): Whether to be in training mode.
        """
        self._training = training
        for agent in self._agents:
            agent.train() if training else agent.eval()

    def create_episode_metrics(self):
        """Create the metrics used during the loop."""
        return Metrics(
            self._agents,
            [("reward", 0), ("episode_length", 0)],
            [("full_episode_length", 0)],
        )

    def run_one_step(self, observation, turn, episode_metrics):
        """Run one step of the training loop.

        Args:
            observation: Current observation that the agent should create an action
                for.
            turn (int): Agent whose turn it is.
            episode_metrics (Metrics): Keeps track of metrics for current episode.
        """
        if self._training:
            self._train_schedule.update()
            self._logger.update_step("train")
            self._run_testing = self._test_schedule.update() or self._run_testing
            self._save_experiment = (
                self._experiment_manager.update_step() or self._save_experiment
            )

    def run_end_step(self, episode_metrics, done):
        """Run the final step of an episode.

        Args:
            episode_metrics (Metrics): Keeps track of metrics for current episode.
            done (bool): Whether this step was terminal.

        """
        return NotImplementedError

    def run_episode(self):
        """Run a single episode of the environment."""

        return NotImplementedError

    def run_training(self):
        """Run the training loop."""
        self.train_mode(True)
        while self._train_schedule.get_value():
            # Run training episode
            if not self._training:
                self.train_mode(True)
            episode_metrics = self.run_episode()
            if self._logger.should_log("train"):
                episode_metrics = episode_metrics.get_flat_dict()
                self._logger.log_metrics(episode_metrics, "train")

            # Run test episodes
            if self._run_testing:
                test_metrics = self.run_testing()
                self._logger.update_step("test")
                self._logger.log_metrics(test_metrics, "test")
                self._run_testing = False

            # Save experiment state
            if self._save_experiment:
                self._experiment_manager.save()
                self._save_experiment = False

        # Run a final test episode and save the experiment.
        test_metrics = self.run_testing()
        self._logger.update_step("test")
        self._logger.log_metrics(test_metrics, "test")
        self._experiment_manager.save()

    def run_testing(self):
        """Run a testing phase."""
        self.train_mode(False)
        aggregated_episode_metrics = self.create_episode_metrics().get_flat_dict()
        episodes = 0
        while episodes <= self._test_episodes:
            episode_metrics = self.run_episode()
            episodes += 1
            for metric, value in episode_metrics.get_flat_dict().items():
                aggregated_episode_metrics[metric] += value / self._test_episodes

        return aggregated_episode_metrics

    def resume(self):
        """Resume a saved experiment."""
        self._experiment_manager.resume()
        self._train_schedule = self._experiment_manager.experiment_state[
            "train_schedule"
        ]
        self._test_schedule = self._experiment_manager.experiment_state["test_schedule"]



from hive.agents import qnets
from hive.agents.agent import Agent
from hive.agents.dqn import DQNAgent
from hive.agents.legal_moves_rainbow import LegalMovesRainbowAgent
from hive.agents.rainbow import RainbowDQNAgent
from hive.agents.random import RandomAgent
from hive.utils.registry import registry

registry.register_all(
    Agent,
    {
        "DQNAgent": DQNAgent,
        "LegalMovesRainbowAgent": LegalMovesRainbowAgent,
        "RainbowDQNAgent": RainbowDQNAgent,
        "RandomAgent": RandomAgent,
    },
)

get_agent = getattr(registry, f"get_{Agent.type_name()}")



from hive.utils.registry import CallableType


class FunctionApproximator(CallableType):
    """A wrapper for callables that produce function approximators.

    For example, :obj:`FunctionApproximator(create_neural_network)` or
    :obj:`FunctionApproximator(MyNeuralNetwork)` where :obj:`create_neural_network` is a
    function that creates a neural network module and :obj:`MyNeuralNetwork` is a
    class that defines your function approximator.

    These wrapped callables can be partially initialized through configuration
    files or command line arguments.
    """

    @classmethod
    def type_name(cls):
        """
        Returns:
            "function"
        """
        return "function"
